from functools import wraps
from typing import TYPE_CHECKING, Any, Callable

import typer

from ._create_di_wrapper import create_di_wrapper

__all__ = ["TyperDI"]


class TyperDI(typer.Typer):
    # we only patch existing methods, so do it silently without affecting type checker

    if not TYPE_CHECKING:

        def __init__(self, *args, **kwargs):
            if "callback" in kwargs:
                kwargs["callback"] = create_di_wrapper(kwargs["callback"])
            super().__init__(*args, **kwargs)

        def callback(self, *args, **kwargs):
            decor = super().callback(*args, **kwargs)
            return wrap_typer_decorator(decor)

        def command(self, *args, **kwargs):
            decor = super().command(*args, **kwargs)
            return wrap_typer_decorator(decor)


def wrap_typer_decorator(decor: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(decor)
    def inner(func):
        # pass wrapper to the typer
        decor(create_di_wrapper(func))

        # return untouched func
        return func

    return inner



import inspect
import sys

if sys.version_info >= (3, 9):
    from typing import Annotated
else:  # pragma: no cover
    from typing_extensions import Annotated

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:  # pragma: no cover
    from typing import Any as TypeAlias


if sys.version_info >= (3, 10):

    def signature(func, eval_str: bool = False):
        return inspect.signature(func, eval_str=eval_str)

else:  # pragma: no cover
    # just don't eval annotations on lower versions
    def signature(func, eval_str: bool = False):
        sig = inspect.signature(func)

        if not eval_str:
            return sig

        def eval_if_str(obj):
            if not isinstance(obj, str):
                return obj
            return eval(obj, sys.modules[func.__module__].__dict__)

        return sig.replace(
            parameters=[
                p.replace(annotation=eval_if_str(p.annotation))
                for p in sig.parameters.values()
            ],
            return_annotation=eval_if_str(sig.return_annotation),
        )



from typing import TYPE_CHECKING, Any, Callable, TypeVar

from .compat import TypeAlias

__all__ = [
    "Depends",
    "DependsType",
]


Callback: TypeAlias = Callable[..., Any]


class DependsType:
    def __init__(self, callback: Callback) -> None:
        self.callback = callback


if TYPE_CHECKING:
    _T = TypeVar("_T")

    def Depends(func: Callable[..., _T]) -> _T:
        ...

else:

    def Depends(func: Callback):
        return DependsType(func)  # type: ignore



from dataclasses import dataclass, field
from inspect import Parameter, Signature
from typing import Dict, Optional, Union

from ._depends import Callback, DependsType
from ._method_builder import MethodBuilder, copy_func_attrs
from .compat import signature

__all__ = [
    "create_di_wrapper",
    "TyperDIError",
]


class TyperDIError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


def create_di_wrapper(func: Callback) -> Callback:
    ctx = _Context()
    _invoke_recursive(ctx, func)
    _sort_params(ctx)
    wrapper = ctx.builder.build()
    copy_func_attrs(wrapper, func)
    return wrapper


@dataclass
class _Context:
    builder: MethodBuilder = field(default_factory=MethodBuilder)
    known_invokes: Dict[Callback, Union[str, None]] = field(default_factory=dict)


def _invoke_recursive(ctx: _Context, func: Callback) -> str:
    """
    Invoke `func` recursively and return the name of the result variable.
    """
    # don't call the same dependency callback twice
    if func in ctx.known_invokes:
        result = ctx.known_invokes[func]
        if result is None:
            # cycle found
            raise TyperDIError(
                f'Found cycle in dependency graph around method "{func.__qualname__}"'
            )

        return result

    ctx.known_invokes[func] = None  # mark this dependency as being processed

    sig = signature(func, eval_str=True)

    kwargs = {}

    for param in sig.parameters.values():
        arg_name = param.name
        depends_type = _parse_dependency(param)
        if depends_type is not None:
            kwargs[arg_name] = _invoke_recursive(ctx, depends_type.callback)
            continue

        # make sure that all name are unique
        if any(param.name == p.name for p in ctx.builder.params):
            raise TyperDIError(
                f"Duplicated parameter name '{param.name}'.\n"
                f"Please, make sure to have unique names in the whole dependency tree."
            )

        # add param to wrapper
        ctx.builder.add_param(
            param.name,
            param.annotation,
            default=param.default,
        )
        kwargs[arg_name] = param.name

    result = ctx.builder.invoke(func, kwargs)
    ctx.known_invokes[func] = result
    return result


def _parse_dependency(param: Parameter) -> Optional[DependsType]:
    if isinstance(param.default, DependsType):
        return param.default

    metadata = getattr(param.annotation, "__metadata__", ())
    for p in metadata:
        if isinstance(p, DependsType):
            return p

    return None


def _sort_params(ctx: _Context) -> None:
    # move params with defaults to the end
    ctx.builder.params.sort(
        key=lambda p: p.default != Signature.empty,
    )



from dataclasses import dataclass
from functools import WRAPPER_ASSIGNMENTS
from inspect import Parameter, Signature
from typing import Any, Dict, List

from ._depends import Callback
from .compat import signature

__all__ = [
    "MethodBuilder",
    "MethodBuilderError",
    "copy_func_attrs",
]


class MethodBuilderError(Exception):
    pass


@dataclass
class ParamInfo:
    name: str
    default: Any = Signature.empty
    annotation: Any = Signature.empty

    def to_parameter(self) -> Parameter:
        return Parameter(
            self.name,
            Parameter.POSITIONAL_OR_KEYWORD,
            default=self.default,
            annotation=self.annotation,
        )


@dataclass
class InvokeInfo:
    callback: Callback
    kwargs: Dict[str, str]  # arguments `k=v` that will be passed to the callback
    result: str  # variable that holds invocation result


_METHOD_TEMPLATE = """\
def func({vars}):
{invokes}
{result}
"""

_INVOKE_TEMPLATE = """\
    {result} = {callback}({args})
"""

_RESULT_TEMPLATE = """\
    return {result}
"""


class MethodBuilder:
    """
    Construct wrapper method in form

    def func({params}):
        {invokes}   # x = foo(y, z)
        return {final_invoke}

    This allows us to bake whole dependency tree to a single function.

    Builder preserves:
     * params type annotations
     * return type annotation and other `wraps` props from `func`
    """

    def __init__(self) -> None:
        self._params: List[ParamInfo] = []
        self._invokes: List[InvokeInfo] = []

    @property
    def params(self) -> List[ParamInfo]:
        return self._params

    def add_param(
        self,
        name: str,
        annotation: type = Signature.empty,
        default: Any = Signature.empty,
    ) -> None:
        self._params.append(
            ParamInfo(name=name, default=default, annotation=annotation)
        )

    def invoke(self, callback: Callback, kwargs: Dict[str, str]) -> str:
        # FIXME: validate that `kwargs.values()` are either in `params` or `calls.result`s
        result = f"__r{len(self._invokes)}"
        self._invokes.append(InvokeInfo(callback, kwargs, result))
        return result

    def build(self) -> Callback:
        program_text = self._format_func_program()
        globs = {f"__cb{idx}": p.callback for idx, p in enumerate(self._invokes)}

        try:
            exec(program_text, globs)
        except Exception as ex:
            raise MethodBuilderError(f"Compilation failed: {ex}\n\n{program_text}")

        func = globs["func"]
        self._update_signature(func)
        return func

    def _format_func_program(self) -> str:
        invokes = []
        for idx, invoke_info in enumerate(self._invokes):
            invokes.append(
                _INVOKE_TEMPLATE.format(
                    result=invoke_info.result,
                    callback=f"__cb{idx}",
                    args=", ".join(f"{k}={v}" for k, v in invoke_info.kwargs.items()),
                )
            )

        if self._invokes:
            # use result of last invokation
            result = _RESULT_TEMPLATE.format(result=self._invokes[-1].result)
        else:
            result = _RESULT_TEMPLATE.format(result="")

        return _METHOD_TEMPLATE.format(
            vars=", ".join(p.name for p in self._params),
            invokes="".join(invokes).rstrip(),
            result=result,
        )

    def _update_signature(self, func: Callback) -> None:
        # take return type from the signature of the last callback
        return_type = Signature.empty
        if self._invokes:
            last_sig = signature(self._invokes[-1].callback, eval_str=True)
            return_type = last_sig.return_annotation

        # setup wrapper signature
        func.__signature__ = Signature(  # type: ignore
            [p.to_parameter() for p in self._params],
            return_annotation=return_type,
        )

        # setup defaults
        func.__defaults__ = tuple(
            param.default
            for param in self._params
            if param.default is not Signature.empty
        )


def copy_func_attrs(wrapper: Callback, func: Callback) -> None:
    # update all except `__annotations__`, to avoid overriding signature
    assigned = set(WRAPPER_ASSIGNMENTS)
    assigned.remove("__annotations__")

    for name in assigned:
        setattr(wrapper, name, getattr(func, name))
